/*
 *   display.c
 *
 *   This file is part of Emu42
 *
 *   Copyright (C) 2004 Christoph Gieelink
 *
 */
#include "pch.h"
#include "resource.h"
#include "Emu42.h"
#include "ops.h"
#include "io.h"
#include "kml.h"

#define DISPLAY_FREQ	16					// display update 1/frequency (1/64) in ms

#define B 0x00000000						// black
#define W 0x00FFFFFF						// white
#define I 0xFFFFFFFF						// ignore

#define LCD_ROW			140					// max. pixel per line (131 or 137)

UINT    nBackgroundX = 0;
UINT    nBackgroundY = 0;
UINT    nBackgroundW = 0;
UINT    nBackgroundH = 0;
UINT    nLcdX = 0;
UINT    nLcdY = 0;
UINT    nLcdZoom = 1;
LPBYTE	pbyLcd;
HDC     hLcdDC = NULL;
HDC     hMainDC = NULL;

static HBITMAP hLcdBitmap;
static HBITMAP hMainBitmap;

static INT  nLcdyZoom,nLcdWidth,nLcdHeight;

static UINT uLcdTimerId = 0;

static DWORD dwKMLColor[64] =				// color table loaded by KML script
{
	W,B,B,B,B,B,B,B,B,B,B,B,B,B,B,B,
	B,B,B,B,B,B,B,B,B,B,B,B,B,B,B,B,
	I,I,I,I,I,I,I,I,I,I,I,I,I,I,I,I,
	I,I,I,I,I,I,I,I,I,I,I,I,I,I,I,I
};

static struct
{
	BITMAPINFOHEADER Lcd_bmih;
	RGBQUAD bmiColors[2];
} bmiLcd =
{
	{sizeof(BITMAPINFOHEADER),0/*x*/,0/*y*/,1,8,BI_RGB,0,0,0,ARRAYSIZEOF(bmiLcd.bmiColors),0}
};

static VOID (*UpdateDisplayMem)(VOID) = NULL;

static VOID UpdateDisplayClamshell(VOID);
static VOID UpdateDisplayPioneer(VOID);

VOID UpdateContrast(BYTE byContrast)
{
	bmiLcd.bmiColors[1] = *(RGBQUAD*)&dwKMLColor[byContrast];    // pixel on  color
	bmiLcd.bmiColors[0] = *(RGBQUAD*)&dwKMLColor[byContrast+32]; // pixel off color
	if ((Chipset.IORam[DSPCTL] & DON) == 0 || I == *(DWORD*)&bmiLcd.bmiColors[0])
		bmiLcd.bmiColors[0] = *(RGBQUAD*)&dwKMLColor[0];

	// update palette information
	EnterCriticalSection(&csGDILock);		// asynchronous display update!
	{
		_ASSERT(hLcdDC);
		SetDIBColorTable(hLcdDC,0,ARRAYSIZEOF(bmiLcd.bmiColors),bmiLcd.bmiColors);
	}
	LeaveCriticalSection(&csGDILock);
	UpdateAnnunciators();					// adjust annunciator color
	return;
}

VOID SetLcdColor(UINT nId, UINT nRed, UINT nGreen, UINT nBlue)
{
	dwKMLColor[nId&0x3F] = ((nRed&0xFF)<<16)|((nGreen&0xFF)<<8)|(nBlue&0xFF);
	return;
}

VOID GetSizeLcdBitmap(INT *pnX, INT *pnY)
{
	*pnX = nLcdWidth  * nLcdZoom;
	*pnY = nLcdHeight * nLcdyZoom;
	return;
}

VOID CreateLcdBitmap(VOID)
{
	// create LCD bitmap
	nLcdyZoom  = nLcdZoom;
	nLcdWidth  = 137;
	nLcdHeight = 32;

	// set display update routine
	UpdateDisplayMem = &UpdateDisplayClamshell;

	if (cCurrentRomType != 'O' && cCurrentRomType != 'Y')
	{
		nLcdWidth  = 131;
		nLcdHeight = 16;

		// set display update routine
		UpdateDisplayMem = &UpdateDisplayPioneer;

		switch (nLcdZoom)
		{
			default:
			case 1 : nLcdyZoom = 2; break;
			case 2 : nLcdyZoom = 3; break;
			case 4 : nLcdyZoom = 6; break;
		}
	}

	bmiLcd.Lcd_bmih.biWidth = LCD_ROW * nLcdZoom;
	bmiLcd.Lcd_bmih.biHeight = -nLcdHeight * nLcdyZoom;
	_ASSERT(hLcdDC == NULL);
	hLcdDC = CreateCompatibleDC(hWindowDC);
	_ASSERT(hLcdDC != NULL);
	hLcdBitmap = CreateDIBSection(hLcdDC,(BITMAPINFO*)&bmiLcd,DIB_RGB_COLORS,(VOID **) &pbyLcd, NULL, 0);
	_ASSERT(hLcdBitmap != NULL);
	// fill with background color at startup
	ZeroMemory(pbyLcd, bmiLcd.Lcd_bmih.biWidth * -bmiLcd.Lcd_bmih.biHeight);
	hLcdBitmap = SelectObject(hLcdDC,hLcdBitmap);
	_ASSERT(hPalette != NULL);
	SelectPalette(hLcdDC, hPalette, FALSE);	// set palette for LCD DC
	RealizePalette(hLcdDC);					// realize palette
	UpdateContrast(Chipset.contrast);
	return;
}

VOID DestroyLcdBitmap(VOID)
{
	// set contrast palette to startup colors
	WORD i = 0;   dwKMLColor[i++] = W;
	while(i < 32) dwKMLColor[i++] = B;
	while(i < 64) dwKMLColor[i++] = I;

	UpdateDisplayMem = NULL;

	if (hLcdDC != NULL)
	{
		// destroy LCD bitmap
		DeleteObject(SelectObject(hLcdDC,hLcdBitmap));
		DeleteDC(hLcdDC);
		hLcdDC = NULL;
		hLcdBitmap = NULL;
	}
	return;
}

BOOL CreateMainBitmap(LPCTSTR szFilename)
{
	HPALETTE hAssertPalette;

	_ASSERT(hWindowDC != NULL);
	hMainDC = CreateCompatibleDC(hWindowDC);
	_ASSERT(hMainDC != NULL);
	if (hMainDC == NULL) return FALSE;		// quit if failed
	hMainBitmap = LoadBitmapFile(szFilename);
	if (hMainBitmap == NULL)
	{
		DeleteDC(hMainDC);
		return FALSE;
	}
	hMainBitmap = SelectObject(hMainDC,hMainBitmap);
	_ASSERT(hPalette != NULL);
	hAssertPalette = SelectPalette(hMainDC,hPalette,FALSE);
	_ASSERT(hAssertPalette != NULL);
	RealizePalette(hMainDC);
	return TRUE;
}

VOID DestroyMainBitmap(VOID)
{
	if (hMainDC != NULL)
	{
		// destroy Main bitmap
		DeleteObject(SelectObject(hMainDC,hMainBitmap));
		DeleteDC(hMainDC);
		hMainDC = NULL;
		hMainBitmap = NULL;
	}
	return;
}

//****************
//*
//* LCD functions
//*
//****************

static __inline VOID WritePixelQuad(BYTE *p, BYTE a)
{
	if (nLcdZoom == 4)
	{
		CONST DWORD dwPattern[] = { 0x00000000, 0x01010101 };

		*(DWORD *)&p[0*LCD_ROW]  = *(DWORD *)&p[4*LCD_ROW]  =
		*(DWORD *)&p[8*LCD_ROW]  = *(DWORD *)&p[12*LCD_ROW] =
		dwPattern[a&1];

		*(DWORD *)&p[16*LCD_ROW] = *(DWORD *)&p[20*LCD_ROW] =
		*(DWORD *)&p[24*LCD_ROW] = *(DWORD *)&p[28*LCD_ROW] =
		dwPattern[(a>>1)&1];

		*(DWORD *)&p[32*LCD_ROW] = *(DWORD *)&p[36*LCD_ROW] =
		*(DWORD *)&p[40*LCD_ROW] = *(DWORD *)&p[44*LCD_ROW] =
		dwPattern[(a>>2)&1];

		*(DWORD *)&p[48*LCD_ROW] = *(DWORD *)&p[52*LCD_ROW] =
		*(DWORD *)&p[56*LCD_ROW] = *(DWORD *)&p[60*LCD_ROW] =
		dwPattern[(a>>3)&1];
	}
	if (nLcdZoom == 2)
	{
		CONST WORD wPattern[] = { 0x0000, 0x0101 };

		*(WORD *)&p[0*LCD_ROW]  =
		*(WORD *)&p[2*LCD_ROW]  = wPattern[a&1];

		*(WORD *)&p[4*LCD_ROW]  =
		*(WORD *)&p[6*LCD_ROW]  = wPattern[(a>>1)&1];

		*(WORD *)&p[8*LCD_ROW]  =
		*(WORD *)&p[10*LCD_ROW] = wPattern[(a>>2)&1];

		*(WORD *)&p[12*LCD_ROW] =
		*(WORD *)&p[14*LCD_ROW] = wPattern[(a>>3)&1];
	}
	if (nLcdZoom == 1)
	{
		p[0*LCD_ROW] = a&1;
		p[1*LCD_ROW] = (a>>1)&1;
		p[2*LCD_ROW] = (a>>2)&1;
		p[3*LCD_ROW] = (a>>3)&1;
	}
	return;
}

static __inline VOID WritePixelRect(BYTE *p, BYTE a)
{
	if (nLcdZoom == 4)
	{
		CONST DWORD dwPattern[] = { 0x00000000, 0x01010101 };

		*(DWORD *)&p[0*LCD_ROW]  = *(DWORD *)&p[4*LCD_ROW]  =
		*(DWORD *)&p[8*LCD_ROW]  = *(DWORD *)&p[12*LCD_ROW] =
		*(DWORD *)&p[16*LCD_ROW] = *(DWORD *)&p[20*LCD_ROW] =
		dwPattern[a&1];

		*(DWORD *)&p[24*LCD_ROW] = *(DWORD *)&p[28*LCD_ROW] =
		*(DWORD *)&p[32*LCD_ROW] = *(DWORD *)&p[36*LCD_ROW] =
		*(DWORD *)&p[40*LCD_ROW] = *(DWORD *)&p[44*LCD_ROW] =
		dwPattern[(a>>1)&1];

		*(DWORD *)&p[48*LCD_ROW] = *(DWORD *)&p[52*LCD_ROW] =
		*(DWORD *)&p[56*LCD_ROW] = *(DWORD *)&p[60*LCD_ROW] =
		*(DWORD *)&p[64*LCD_ROW] = *(DWORD *)&p[68*LCD_ROW] =
		dwPattern[(a>>2)&1];

		*(DWORD *)&p[72*LCD_ROW] = *(DWORD *)&p[76*LCD_ROW] =
		*(DWORD *)&p[80*LCD_ROW] = *(DWORD *)&p[84*LCD_ROW] =
		*(DWORD *)&p[88*LCD_ROW] = *(DWORD *)&p[92*LCD_ROW] =
		dwPattern[(a>>3)&1];
	}
	if (nLcdZoom == 2)
	{
		CONST WORD wPattern[] = { 0x0000, 0x0101 };

		*(WORD *)&p[0*LCD_ROW]  =
		*(WORD *)&p[2*LCD_ROW]  =
		*(WORD *)&p[4*LCD_ROW]  = wPattern[a&1];

		*(WORD *)&p[6*LCD_ROW]  =
		*(WORD *)&p[8*LCD_ROW]  =
		*(WORD *)&p[10*LCD_ROW] = wPattern[(a>>1)&1];

		*(WORD *)&p[12*LCD_ROW] =
		*(WORD *)&p[14*LCD_ROW] =
		*(WORD *)&p[16*LCD_ROW] = wPattern[(a>>2)&1];

		*(WORD *)&p[18*LCD_ROW] =
		*(WORD *)&p[20*LCD_ROW] =
		*(WORD *)&p[22*LCD_ROW] = wPattern[(a>>3)&1];
	}
	if (nLcdZoom == 1)
	{
		p[0*LCD_ROW] = p[1*LCD_ROW] = a&1;
		p[2*LCD_ROW] = p[3*LCD_ROW] = (a>>1)&1;
		p[4*LCD_ROW] = p[5*LCD_ROW] = (a>>2)&1;
		p[6*LCD_ROW] = p[7*LCD_ROW] = (a>>3)&1;
	}
	return;
}

static VOID UpdateDisplayClamshell(VOID)
{
	DWORD d;
	INT   x0,y0;
	BYTE  *p;

	// scan complete display area of master
	for (d = MDISP_BEGIN; d <= MDISP_END; ++d)
	{
		// calculate x/y coordinates
		y0 = (d&0x7)*4;
		x0 = (SDISP_END-SDISP_BEGIN+1+d)>>3;

		// calculate memory position in LCD bitmap
		p = pbyLcd + (y0*LCD_ROW*nLcdyZoom + x0)*nLcdZoom;

		// write pixel zoom independent
		WritePixelQuad(p,Chipset.IORam[d]);
	}

	// scan complete display area of slave
	for (d = SDISP_BEGIN; d <= SDISP_END; ++d)
	{
		// calculate x/y coordinates
		y0 = ((d-SDISP_BEGIN)&0x7)*4;
		x0 = (d-SDISP_BEGIN)>>3;

		// calculate memory position in LCD bitmap
		p = pbyLcd + (y0*LCD_ROW*nLcdyZoom + x0)*nLcdZoom;

		// write pixel zoom independent
		WritePixelQuad(p,(BYTE)(ChipsetS.bLcdSyncErr ? 0xF : ChipsetS.IORam[d]));
	}
	return;
}

static VOID UpdateDisplayPioneer(VOID)
{
	DWORD d;
	INT   x0,y0;
	BYTE  *p;

	// scan complete display area
	for (d = DISP_BEGIN; d <= DISP_END; ++d)
	{
		// calculate x/y coordinates
		y0 = (d&0x3)*4;
		x0 = 66*((d>>2)&1)+(d>>3);

		// calculate memory position in LCD bitmap
		p = pbyLcd + (y0*LCD_ROW*nLcdyZoom + x0)*nLcdZoom;

		// write pixel zoom independent
		WritePixelRect(p,Chipset.IORam[d]);
	}
	return;
}

static VOID UpdateMainDisplay(VOID)
{
	// illegal zoom factor
	_ASSERT(nLcdZoom == 1 || nLcdZoom == 2 || nLcdZoom == 4);

	_ASSERT(UpdateDisplayMem);				// function not initialized
	UpdateDisplayMem();						// update display memory

	EnterCriticalSection(&csGDILock);		// solving NT GDI problems
	{
		BitBlt(hWindowDC, nLcdX, nLcdY, nLcdWidth * nLcdZoom, -bmiLcd.Lcd_bmih.biHeight,
			   hLcdDC, 0, 0, SRCCOPY);
		GdiFlush();
	}
	LeaveCriticalSection(&csGDILock);
	return;
}

VOID UpdateAnnunciators(VOID)
{
	const DWORD dwAnnAddrP[] = { LA_ALL, LA_UPDOWN, LA_SHIFT, LA_PRINTER, LA_BUSY, LA_BAT, LA_G, LA_RAD };
	const DWORD dwAnnAddrO[] = { SLA_ALL, SLA_HALT, SLA_SHIFT, SLA_ALPHA, SLA_BUSY, SLA_BAT, SLA_RAD, SLA_PRINTER };

	DWORD  const *dwAnnAddr;
	LPBYTE pbySource;

	INT   i,j,nCount;

	BOOL bAnnOn = (Chipset.IORam[DSPCTL] & DON) != 0;

	if (Chipset.bSlave)						// has slave controller
	{
		dwAnnAddr = dwAnnAddrO;
		pbySource = ChipsetS.IORam;			// slave controller

		bAnnOn = bAnnOn && (ChipsetS.IORam[DSPCTL] & DON) != 0;
	}
	else
	{
		dwAnnAddr = dwAnnAddrP;
		pbySource = Chipset.IORam;			// master controller
	}

	// check all annuncators
	for (i = 1; i < ARRAYSIZEOF(dwAnnAddrP); ++i)
	{
		nCount = 0;

		if (bAnnOn)							// annunciators on
		{
			DWORD dwAnn = Npack(pbySource+dwAnnAddr[i],8) ^ Npack(pbySource+dwAnnAddr[0],8);

			// count the number of set bits
			for (;dwAnn != 0; dwAnn >>= 1)
			{
				if ((dwAnn & 0x1) != 0) ++nCount;
			}
		}

		// contrast table entry of annunciator
		j = Chipset.contrast + nCount - 16;
		if (j < 0)  j = 0;
		if (j > 31) j = 31;

		DrawAnnunciator(i, nCount > 0, dwKMLColor[j]);
	}
	return;
}

static VOID CALLBACK LcdProc(UINT uEventId, UINT uMsg, DWORD dwUser, DWORD dw1, DWORD dw2)
{
	UpdateMainDisplay();
	return;
	UNREFERENCED_PARAMETER(uEventId);
	UNREFERENCED_PARAMETER(uMsg);
	UNREFERENCED_PARAMETER(dwUser);
	UNREFERENCED_PARAMETER(dw1);
	UNREFERENCED_PARAMETER(dw2);
}

VOID StartDisplay(VOID)
{
	if (uLcdTimerId)						// LCD update timer running
		return;								// -> quit

	if (Chipset.IORam[DSPCTL]&DON)			// display on?
	{
		UpdateAnnunciators();				// switch on annunciators
		VERIFY(uLcdTimerId = timeSetEvent(DISPLAY_FREQ,0,(LPTIMECALLBACK)&LcdProc,0,TIME_PERIODIC));
	}
	return;
}

VOID StopDisplay(VOID)
{
	// dummy region, WM_PAINT redraw always complete display area
	RECT Rect = { nLcdX, nLcdY, nLcdX+1, nLcdY+1 };

	if (uLcdTimerId == 0)					// timer stopped
		return;								// -> quit

	timeKillEvent(uLcdTimerId);				// stop display update
	uLcdTimerId = 0;						// set flag display update stopped

	ZeroMemory(pbyLcd, bmiLcd.Lcd_bmih.biWidth * -bmiLcd.Lcd_bmih.biHeight);
	InvalidateRect(hWnd,&Rect,FALSE);
	return;
}

VOID ResizeWindow(VOID)
{
	RECT rectWindow;
	RECT rectClient;

	if (hWnd == NULL) return;				// return if window closed

	rectWindow.left   = 0;
	rectWindow.top    = 0;
	rectWindow.right  = nBackgroundW;
	rectWindow.bottom = nBackgroundH;
	AdjustWindowRect(&rectWindow, WS_CAPTION|WS_SYSMENU|WS_MINIMIZEBOX|WS_OVERLAPPED, TRUE);
	SetWindowPos (hWnd, (HWND)NULL, 0, 0,
		rectWindow.right  - rectWindow.left,
		rectWindow.bottom - rectWindow.top,
		SWP_NOMOVE | SWP_NOZORDER);
	GetClientRect(hWnd, &rectClient);
	AdjustWindowRect(&rectClient, WS_CAPTION|WS_SYSMENU|WS_MINIMIZEBOX|WS_OVERLAPPED, TRUE);
	if (rectClient.bottom < rectWindow.bottom)
	{
		rectWindow.bottom += (rectWindow.bottom - rectClient.bottom);
		SetWindowPos (hWnd, (HWND)NULL, 0, 0,
			rectWindow.right  - rectWindow.left,
			rectWindow.bottom - rectWindow.top,
			SWP_NOMOVE | SWP_NOZORDER);
	}

	_ASSERT(hWindowDC);						// move destination window
	SetWindowOrgEx(hWindowDC, nBackgroundX, nBackgroundY, NULL);
	InvalidateRect(hWnd,NULL,TRUE);
	return;
}
